Energized = LibStub("AceAddon-3.0"):NewAddon("Energized", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
Energized.MINOR_VERSION = tonumber(("$Revision: 33 $"):match("%d+"))
local L = LibStub("AceLocale-3.0"):GetLocale("Energized")
local E = Energized
local qtip = LibStub("LibQTip-1.0")
local talents = LibStub("LibGroupTalents-1.0")
local ldb = LibStub:GetLibrary("LibDataBroker-1.1")
local dataobj = ldb:NewDataObject("Energized", {
 type = "data source",
 text = L["Loading"],
 icon = [[Interface\AddOns\Energized\Icons\blank]],
 label = "Energized",
 tocname = "Energized",
})
local dbicon = LibStub("LibDBIcon-1.0")



--[[
      VARIABLES ----------------------------------------------------------------------------------------------------------------
  ]]

E.isEnabled = false
E.inCombat = false

local players = {}
local classes = {}
local aura = {
	me = {
		spellName = {},
		spellId = {},
		expiration = {},
		unitCaster = {},
	},
	pet = {
		spellName = {},
		spellId = {},
		expiration = {},
		unitCaster = {},
	},
}
local items = {}

local otherIssuesA = {
	me = {
	},
	pet = {
	},
}
local otherIssuesB = {
	me = {
	},
	pet = {
	},
}
local gearA = {}
local gearB = {}

local isHeroic = false
local partyFormat
local updateSource
local playerName
local hideScheduled = false
local isHidden = false
local hideTimer

local POLLING_TIME = 10.0
local UPDATE_THROTTLE = 2.0

local CLASS_LIST = {
	"HUNTER",
	"WARRIOR",
	"ROGUE",
	"SHAMAN",
	"PRIEST",
	"WARLOCK",
	"DEATHKNIGHT",
	"PALADIN",
	"MAGE",
	"DRUID",
}

local invSlots = { "HeadSlot", "NeckSlot", "ShoulderSlot", "BackSlot", "ChestSlot", "WristSlot", "HandsSlot", "WaistSlot", "LegsSlot", "FeetSlot", "Finger0Slot", "Finger1Slot", "Trinket0Slot", "Trinket1Slot", "MainHandSlot", "SecondaryHandSlot", "RangedSlot" }

local red = { r = 255, g = 51, b = 51 }
local blue = { r = 102, g = 102, b = 255 }
local green = { r = 51, g = 255, b = 51 }
local yellow = { r = 255, g = 255, b = 51 }
local gray = { r = 128, g = 128, b = 128 }




--[[
      UPVALUES -----------------------------------------------------------------------------------------------------------------
  ]]
  
local _G = getfenv(0)
local pairs = _G.pairs
local strlower = _G.strlower
local strupper = _G.strupper
local format = _G.format

local UnitInRaid = _G.UnitInRaid
local UnitInParty = _G.UnitInParty
local UnitClass = _G.UnitClass
local UnitName = _G.UnitName
local UnitAura = _G.UnitAura
local UnitExists = _G.UnitExists
local UnitIsConnected = _G.UnitIsConnected
local GetNumRaidMembers = _G.GetNumRaidMembers
local GetNumPartyMembers = _G.GetNumPartyMembers
local GetRaidRosterInfo = _G.GetRaidRosterInfo
local UnitIsDeadOrGhost = _G.UnitIsDeadOrGhost
local UnitIsFeignDeath = _G.UnitIsFeignDeath
local GetBattlefieldStatus = _G.GetBattlefieldStatus
local GetPetActionInfo = _G.GetPetActionInfo
local GetSubZoneText = _G.GetSubZoneText
local GetInstanceDifficulty = _G.GetInstanceDifficulty
local GetInstanceInfo = _G.GetInstanceInfo
local IsInInstance = _G.IsInInstance
local GetSpellInfo = _G.GetSpellInfo
local GetItemInfo = _G.GetItemInfo
local GetPetHappiness = _G.GetPetHappiness
local IsMounted = _G.IsMounted
local GetInventoryItemLink = _G.GetInventoryItemLink
local GetInventorySlotInfo = _G.GetInventorySlotInfo
local GetInventoryItemQuality = _G.GetInventoryItemQuality
local GetInventoryItemBroken = _G.GetInventoryItemBroken
local GetInventoryItemDurability = _G.GetInventoryItemDurability
local GetInventoryItemID = _G.GetInventoryItemID
local GetContainerNumSlots = _G.GetContainerNumSlots
local GetContainerItemID = _G.GetContainerItemID
local GetContainerItemInfo = _G.GetContainerItemInfo

local MAX_BATTLEFIELD_QUEUES = _G.MAX_BATTLEFIELD_QUEUES
local NUM_BAG_SLOTS = _G.NUM_BAG_SLOTS



--[[
      INITIALIZATION -----------------------------------------------------------------------------------------------------------
  ]]

function E:OnInitialize()
	self.class = select(2, UnitClass("player"))
	playerName = UnitName("player")
end


do
	local isInitialized = false
	
	function E:OnEnable()
		if isInitialized then return end
		isInitialized = true
	
		self:CheckForPetClass()
		self:PrepareDefaults()
		self:PrepareClassDefaults()
		
		self.db = LibStub("AceDB-3.0"):New("EnergizedDB", self.defaults, "Default")
		self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
		self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
		self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
		self.db.RegisterCallback(self, "OnProfileDeleted", "OnProfileChanged")
		
		self:Cleanup()

		local LibDualSpec = LibStub("LibDualSpec-1.0")
		LibDualSpec:EnhanceDatabase(self.db, "Energized")

		self:RegisterChatCommand("energize", "ChatCommand")
		self:RegisterChatCommand("energized", "ChatCommand")
		
		self.db.profile._currentRuleType = nil
		
		dbicon:Register("Energized", dataobj, self.db.profile.minimap)

		E.isEnabled = true
		self:UpdateRaid("INITIALIZING")
		
		talents.RegisterCallback(self, "LibGroupTalents_Update", "OnYourSpecChanged")
		talents.RegisterCallback(self, "LibGroupTalents_UpdateComplete", "OnGroupTalentsUpdate")
		talents.RegisterCallback(self, "LibGroupTalents_Add", "OnGroupTalentsUpdate")
		talents.RegisterCallback(self, "LibGroupTalents_Remove", "OnGroupTalentsUpdate")

		if self.db.profile.debugMode then
			Energized.players = players
			Energized.classes = classes
			Energized.aura = aura
			Energized.items = items
		end
		
		self:RegisterConstantEvents()
		self:RegisterOutOfCombatEvents()
		self:ScheduleRepeatingTimer("OnTimer", POLLING_TIME)

		self:Debug(L["Loaded (running in debug mode)"])
	end
end


function E:OnDisable()
	E.isEnabled = false
	self:Debug(L["Disabled"])
end


function E:ChatCommand(input)
	if strlower(input) == "debug" then
		if self.db.profile.debugMode then
			self.db.profile.debugMode = false
			self:Print(L["Debug mode OFF"])
		else
			self.db.profile.debugMode = true
			self:Print(L["Debug mode ON"])
		end
	else
		LoadAddOn("Energized_Options")
		InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
	end
end



--[[
      UTILITIES ----------------------------------------------------------------------------------------------------------------
  ]]

function E:tcopy(to, from)
 for k, v in pairs(from) do
  if type(v) == "table" then
   to[k] = {}
   E:tcopy(to[k], v)
  else
   to[k] = v
  end
 end
end


local function colorize(s, color)
	if color and s then
		return string.format("|cff%02x%02x%02x%s|r", color.r or 255, color.g or 255, color.b or 255, s)
	else
		return s
	end
end


function E:Debug(s, ...)
	if self.db.profile.debugMode then self:Print(format(s, ...)) end
end


local function strstarts(s, start)
 return strsub(s, 1, strlen(start)) == start
end


local function strends(s, ending)
 return ending == "" or strsub(s, -strlen(ending)) == ending
end


local compare_funcs = {
	["EXACT"] = function(a, b) return a == b end,
	["PARTIAL"] = function(a, b) return strfind(a, b) end,
	["STARTSWITH"] = function(a, b) return strstarts(a, b) end,
	["ENDSWITH"] = function(a, b) return strends(a, b) end,
}



--[[
      DATA BROKER OBJECT -------------------------------------------------------------------------------------------------------
  ]]

do
	local tooltip, tooltip2
	local frame, frame2
	local header1added, header2added, header3added, header4added
	local spacer = "   "

	function dataobj.OnEnter(self)
		if not E:ShouldEnable() then return end
		hideScheduled = false
		E:CancelTimer(hideTimer, true)
		hideTimer = nil
		E:ShowFeed()
		frame = self
		E:ShowTooltip()
		E:Update("LDB ONENTER")
	end


	function dataobj.OnLeave(self)
		E:Update("LDB ONLEAVE")
	end
	
	
	function E:InTooltip()
		return qtip:IsAcquired("EnergizedTooltip")
	end


	function dataobj:OnClick(button)
		LoadAddOn("Energized_Options")
		InterfaceOptionsFrame_OpenToCategory(E.optionsFrame)
	end
	
	
	local function addHeader1()
		if not header1added then
			header1added = true
			tooltip:AddLine(colorize(L["You"], blue))
		end
	end


	local function addHeader2()
		if not header2added then
			header2added = true
			if header1added then tooltip:AddSeparator() end
			tooltip:AddLine(colorize(L["Your pet"], blue))
		end
	end


	local function addHeader3()
		if not header3added then
			header3added = true
			if header1added or header2added then tooltip:AddSeparator() end
			tooltip:AddLine(colorize(L["Equipment"], blue))
		end
	end
	
	
	function E:OpenSubTooltip(frame)
		if qtip:IsAcquired("EnergizedSubTooltip") and tooltip2 then
			qtip:Release(tooltip2)
			tooltip2 = nil
		end
		tooltip2 = qtip:Acquire("EnergizedSubTooltip", 3, "LEFT")
		tooltip2:ClearAllPoints()
		tooltip2:SetClampedToScreen(true)
		tooltip2:SetPoint("LEFT", frame, "RIGHT", 10, 0)
		return tooltip2
	end


	function E:ShowTooltip()
		if not E:ShouldEnable() then return end
		if qtip:IsAcquired("EnergizedTooltip") and tooltip then
			tooltip:Clear()
		else
			tooltip = qtip:Acquire("EnergizedTooltip", 4, "LEFT") -- intentionally one column more than we need to avoid text clipping
		end
		tooltip:AddHeader("Energized")
		tooltip:AddSeparator()
		
		header1added = false
		header2added = false
		header3added = false
		
		-- Your buffs
		for _, group in pairs(self.db.profile.groups) do
			if group._result then
				if (group._result == "Low" and self.db.profile.lowWarn) or group._result ~= "Low" then
					addHeader1()
					local result = L[group._result]
					if group._result == "Expected" or group._result == "Unexpected" then result = colorize(result, red)
					elseif group._result == "Low" or group._result == "Missing" then result = colorize(result, yellow)
					end
					local lastBy, lastTime = self:FindLastBufferMe(group)
					local last = ""
					if lastBy and lastTime > 0 and (GetTime() - lastTime) < 7200 then
						if (self.db.profile.showYourselfAsRecentCaster and lastBy == playerName) or lastBy ~= playerName then
							if self.db.profile.recentCaster == 3 then
								last = colorize(format(L["%s (%d |4minute:minutes; ago)"], lastBy, (GetTime() - lastTime) / 60), gray)
							elseif self.db.profile.recentCaster == 2 then
								last = colorize(lastBy, gray)
							end
						end
					end
					local line = tooltip:AddLine(spacer..result, group.name, last)
					-- tooltip:SetLineScript(line, "OnEnter", function(frame)
						-- self:OpenSubTooltip(frame)
						-- tooltip2:AddLine("test")
						-- tooltip2:Show()
					-- end)
					-- tooltip:SetLineScript(line, "OnLeave", function()
						-- if tooltip2 then
							-- qtip:Release(tooltip2)
							-- tooltip2 = nil
						-- end
					-- end)
				end
			end
		end
		
		-- Your additional issues
		for i = 1, #otherIssuesA.me do
			addHeader1()
			tooltip:AddLine(spacer..otherIssuesA.me[i], otherIssuesB.me[i])
		end

		-- Pet buffs
		if self.db.profile.petBuffs and E.isPetClass and UnitExists("playerpet") then
			for _, group in pairs(self.db.profile.groups) do
				if group._result_pet then
					if (group._result_pet == "Low" and self.db.profile.lowWarn) or group._result_pet ~= "Low" then
						addHeader2()
						local result = L[group._result_pet]
						if group._result_pet == "Expected" or group._result_pet == "Unexpected" then result = colorize(result, red)
						elseif group._result_pet == "Low" then result = colorize(result, yellow)
						end
						local lastBy, lastTime = self:FindLastBufferPet(group)
						local last
						if lastBy and lastTime > 0 and (GetTime() - lastTime) < 7200 then
							if (self.db.profile.showYourselfAsRecentCaster and lastBy == playerName) or lastBy ~= playerName then
								if self.db.profile.recentCaster == 3 then
									last = colorize(format(L["%s (%d |4minute:minutes; ago)"], lastBy, (GetTime() - lastTime) / 60), gray)
								elseif self.db.profile.recentCaster == 2 then
									last = colorize(lastBy, gray)
								end
							end
						end
						tooltip:AddLine(spacer..result, group.name, last)
					end
				end
			end
		end
		
		-- Pet additional issues
		for i = 1, #otherIssuesA.pet do
			addHeader2()
			tooltip:AddLine(spacer..otherIssuesA.pet[i], otherIssuesB.pet[i])
		end
		
		-- Gear issues
		for i = 1, #gearA do
			addHeader3()
			tooltip:AddLine(spacer..gearA[i], gearB[i])
		end

		-- No issues
		if not header1added and not header2added and not header3added then
			tooltip:AddLine(colorize(L["You are energized!"], green))
		end
				
		tooltip:SetAutoHideDelay(0.01, frame)
		tooltip:SmartAnchorTo(frame)
		tooltip:UpdateScrolling()
		tooltip:Show()
	end
end



--[[
      EVENTS -------------------------------------------------------------------------------------------------------------------
  ]]

function E:OnProfileChanged(event, database, newProfileKey)
	self:Debug("Profile changed. Reinitializing.")
	self:CheckForPetClass()
	self:PrepareClassDefaults()
	self.db:RegisterDefaults(self.defaults)
	if self.SetupGroupList then self:SetupGroupList() end
	self:UpdateRaid()
end


function E:UnregisterOutOfCombatEvents()
	self:UnregisterEvent("RAID_ROSTER_UPDATE", "UpdateRaid")
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED", "UpdateParty")
	self:UnregisterEvent("PARTY_MEMBER_DISABLE", "UpdateRaid")
	self:UnregisterEvent("PARTY_MEMBER_ENABLE", "UpdateRaid")
	self:UnregisterEvent("ZONE_CHANGED_NEW_AREA", "OnZone")
	self:UnregisterEvent("ZONE_CHANGED", "OnZone")
	self:UnregisterEvent("ZONE_CHANGED_INDOORS", "OnZone")
	self:UnregisterEvent("PARTY_INVITE_REQUEST", "Cleanup")
	self:UnregisterEvent("PARTY_CONVERTED_TO_RAID", "Cleanup")
	self:UnregisterEvent("PLAYER_ENTERING_BATTLEGROUND", "Cleanup")
	self:UnregisterEvent("UNIT_AURA", "OnAura")
	self:UnregisterEvent("UNIT_PET", "Update")
	self:UnregisterEvent("UNIT_INVENTORY_CHANGED", "Update")
	self:UnregisterEvent("UPDATE_INVENTORY_ALERTS", "Update")
	self:UnregisterEvent("READY_CHECK", "OnReadyCheck")
end


function E:RegisterOutOfCombatEvents()
	self:RegisterEvent("RAID_ROSTER_UPDATE", "UpdateRaid")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "UpdateParty")
	self:RegisterEvent("PARTY_MEMBER_DISABLE", "UpdateRaid")
	self:RegisterEvent("PARTY_MEMBER_ENABLE", "UpdateRaid")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "OnZone")
	self:RegisterEvent("ZONE_CHANGED", "OnZone")
	self:RegisterEvent("ZONE_CHANGED_INDOORS", "OnZone")
	self:RegisterEvent("PARTY_INVITE_REQUEST", "Cleanup")
	self:RegisterEvent("PARTY_CONVERTED_TO_RAID", "Cleanup")
	self:RegisterEvent("PLAYER_ENTERING_BATTLEGROUND", "Cleanup")
	self:RegisterEvent("UNIT_AURA", "OnAura")
	self:RegisterEvent("UNIT_PET", "Update")
	self:RegisterEvent("UNIT_INVENTORY_CHANGED", "Update")
	self:RegisterEvent("UPDATE_INVENTORY_ALERTS", "Update")
	self:RegisterEvent("READY_CHECK", "OnReadyCheck")
end


function E:RegisterConstantEvents()
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "OnEvent")
	self:RegisterEvent("PLAYER_REGEN_ENABLED", "OnEvent")
	self:RegisterEvent("PLAYER_LOGOUT", "Cleanup")
end


function E:OnEvent(event, ...)
	if event == "PLAYER_REGEN_DISABLED" then
		self:UnregisterOutOfCombatEvents()
		self.inCombat = true
	elseif event == "PLAYER_REGEN_ENABLED" then
		self:RegisterOutOfCombatEvents()
		self.inCombat = false
 end
end


function E:OnAura(event, unit)
	if unit == "player" or unit == "playerpet" or unit == "pet" then
		updateSource = event
		self:Update()
	end
end


function E:OnTimer()
	updateSource = "TIMER"
	if E.inCombat then self:UpdateRaid("TIMER") else self:Update() end
end


function E:OnZone(event)
	updateSource = event
	self:UpdateRaid()
end


function E:UpdateRaid(event)
	if UnitInRaid("player") then
		if event then updateSource = event end
		table.wipe(players)
		players[UnitName("player")] = self.class
		local myZone = GetRealZoneText()
		for i = 1, GetNumRaidMembers() do
			local playerName, rank, subgroup, level, class, fileName, zone, online, isDead, role, isML = GetRaidRosterInfo(i)
			if playerName and online and not isDead and (zone == myZone or not zone) and (not UnitIsDeadOrGhost(playerName) or UnitIsFeignDeath(playerName)) and UnitIsConnected(playerName) then
				players[playerName] = select(2, UnitClass(playerName))
			end
		end
		self:CountClasses()
		self:Update()
	elseif GetNumPartyMembers() > 0 then
		self:UpdateParty(event)
	else
		self:UpdateSolo(event)
	end
end


function E:UpdateParty(event)
	if not UnitInRaid("player") and GetNumPartyMembers() > 0 then
		if event then updateSource = event end
		table.wipe(players)
		players[UnitName("player")] = self.class
		for i = 1, GetNumPartyMembers() do
			local name = UnitName("party"..i)
			if name then
				if (not UnitIsDeadOrGhost(name) or UnitIsFeignDeath(name)) and UnitIsConnected(name) then
					players[name] = select(2, UnitClass(name))
				end
			end
		end
		self:CountClasses()
		self:Update()
	elseif not UnitInRaid("player") then
		self:UpdateSolo(event)
	end
end


function E:UpdateSolo(event)
	if not UnitInRaid("player") and GetNumPartyMembers() <= 0 then
		if event then updateSource = event end
		table.wipe(players)
		players[UnitName("player")] = self.class
		self:CountClasses()
		self:Update()
	end
end


function E:Cleanup()
	if self.db then
		self:Debug("Cleaning up temporary information")
		for k, v in pairs(self.db.profile.groups) do
			v._status = nil
			v._status_pet = nil
			v._result = nil
			v._result_pet = nil
			for kk, vv in pairs(v.items) do
				vv._lastProvidedBy = nil
				vv._lastSeen = nil
				vv._lastProvidedBy_pet = nil
				vv._lastSeen_pet = nil
				vv._status = nil
				vv._status_pet = nil
			end
		end
	end
end


function E:OnYourSpecChanged(e, guid, unit, newSpec, n1, n2, n3, oldSpec, o1, o2, o3)
	if not E.inCombat then
		if unit == "player" then
			self:Debug("Your spec changed. Recreating default configuration.")
			self:CheckForPetClass()
			self:PrepareClassDefaults()
			self.db:RegisterDefaults(self.defaults)
			if self.SetupGroupList then self:SetupGroupList() end
		end
		updateSource = "LIBGROUPTALENTS: YOUR SPEC CHANGED"
		self:Update()
	end
end


function E:OnGroupTalentsUpdate()
	if not E.inCombat then
		updateSource = "LIBGROUPTALENTS"
		self:Update()
	end
end


function E:OnReadyCheck()
	self:UpdateRaid()
	hideScheduled = false
	self:ShowFeed()
end



--[[
      PROCESSING ---------------------------------------------------------------------------------------------------------------
  ]]


do
	local lastUpdate
	local updateScheduled = false
	local scanNum = 0
	
	function E:_Update()
		local status, numIssues

		updateScheduled = false
		
		if not self:ShouldEnable() then
			self:SetStatus("blank", "")
		else
		
			scanNum = scanNum + 1
			local updateReason = ""
			if updateSource then updateReason = "("..updateSource..")" end
			updateSource = nil
			self:Debug("----> Scanning %d "..updateReason.."  last was %.2f seconds ago", scanNum, GetTime() - (lastUpdate or GetTime()))
			lastUpdate = GetTime()

			-- Get instance difficulty and party format
			do
				local diff = GetInstanceDifficulty()
				local name, type, difficultyIndex, difficultyName, maxPlayers, dynamicDifficulty, isDynamic = GetInstanceInfo()
				local inInstance, instanceType = IsInInstance()
				if isDynamic then
					if dynamicDifficulty == 1 then isHeroic = true else isHeroic = false end
				else
					if diff == 3 or diff == 4 or (diff == 2 and maxPlayers == 5) then isHeroic = true else isHeroic = false end
				end
				partyFormat = "NONE"
				if UnitInRaid("player") then partyFormat = "RAID_UNKNOWN" end
				if inInstance then
					if maxPlayers > 10 then partyFormat = "RAID_25"
					elseif maxPlayers == 10 then partyFormat = "RAID_10"
					elseif maxPlayers == 5 then partyFormat = "PARTY"
					end
				end
			end
			
			--self:Debug("isHeroic: %s    partyFormat: %s", isHeroic and "true" or "false", partyFormat)
			
			-- Scan auras
			do
				table.wipe(aura.me.spellName)
				table.wipe(aura.me.spellId)
				table.wipe(aura.me.expiration)
				table.wipe(aura.me.unitCaster)
			
				local name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId
				for i = 1, 40 do
					name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId = UnitAura("player", i)
					if name then
						name = strupper(name)
						table.insert(aura.me.spellName, name)
						table.insert(aura.me.spellId, spellId)
						table.insert(aura.me.expiration, expirationTime)
						table.insert(aura.me.unitCaster, unitCaster)
					else
						break
					end
				end

				table.wipe(aura.pet.spellName)
				table.wipe(aura.pet.spellId)
				table.wipe(aura.pet.expiration)
				table.wipe(aura.pet.unitCaster)
			
				local name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId
				for i = 1, 40 do
					name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId = UnitAura("playerpet", i)
					if name then
						name = strupper(name)
						table.insert(aura.pet.spellName, name)
						table.insert(aura.pet.spellId, spellId)
						table.insert(aura.pet.expiration, expirationTime)
						table.insert(aura.pet.unitCaster, unitCaster)
					else
						break
					end
				end
			end

			-- Scan inventory
			table.wipe(items)
			if self.db.profile.itemWarn then
				for i = 0, NUM_BAG_SLOTS do
					local numSlots = GetContainerNumSlots(i)
					if numSlots then
						for ii = 1, numSlots do
							local id = GetContainerItemID(i, ii)
							if id then
								local qty = select(2, GetContainerItemInfo(i, ii))
								if not items[id] then items[id] = 0 end
								items[id] = items[id] + qty
							end
						end
					end
				end
			end
			
			-- Evaluate all rule groups
			status, numIssues = self:EvaluateAllGroups()
			
			-- Check other issues
			table.wipe(otherIssuesA.me)
			table.wipe(otherIssuesB.me)
			table.wipe(otherIssuesA.pet)
			table.wipe(otherIssuesB.pet)
			
			-- Pet checks
			if E.isPetClass then
				-- Pet exists
				if not UnitExists("pet") and not IsMounted() then
					table.insert(otherIssuesA.pet, colorize(L["Missing"], red))
					table.insert(otherIssuesB.pet, L["Pet"])
				elseif UnitExists("pet") then
					-- Happiness
					if E.class == "HUNTER" and GetPetHappiness() ~= 3 then
						table.insert(otherIssuesA.pet, colorize(L["Unhappy"], red))
						table.insert(otherIssuesB.pet, L["Pet"])
					end
					-- Attack mode
					if self.db.profile.petMode ~= "NONE" and self:GetPetMode() ~= self.db.profile.petMode then
						table.insert(otherIssuesA.pet, colorize(L["Expected"], red))
						if self.db.profile.petMode == "PET_MODE_PASSIVE" then
							table.insert(otherIssuesB.pet, L["Passive"])
						else
							table.insert(otherIssuesB.pet, L["Defensive"])
						end
					end
				end
			end
			
			-- Check equipment
			table.wipe(gearA)
			table.wipe(gearB)
			local missingSlots = 0
			local lowSlots = 0
			local brokenSlots = 0
			local twoHander = false
			local itemLink = GetInventoryItemLink("player", GetInventorySlotInfo("MainHandSlot"))
			if itemLink and select(9, GetItemInfo(itemLink)) == "INVTYPE_2HWEAPON" then twoHander = true end
			for _, slot in pairs(invSlots) do
				local slotId = select(1, GetInventorySlotInfo(slot))
				local itemId = GetInventoryItemID("player", slotId)
				if itemId and self.db.profile.doNotEquip[itemId] then
					status = "red"
					table.insert(gearA, colorize(L["Unexpected"], red))
					table.insert(gearB, GetInventoryItemLink("player", slotId))
				elseif not GetInventoryItemQuality("player", slotId) then
					if slot ~= "SecondaryHandSlot" or not twoHander then
						missingSlots = missingSlots + 1
					end
				elseif GetInventoryItemBroken("player", slotId) then
					brokenSlots = brokenSlots + 1
				else
					local value, max = GetInventoryItemDurability(slotId)
					local percent = 1
					if max and max ~= 0 and value then percent = value / max end
					if percent <= self.db.profile.minDurability then
						lowSlots = lowSlots + 1
					end
				end
			end
			if brokenSlots > 0 then
			 status = "red"
			 table.insert(gearA, colorize(L["Broken"], red))
			 table.insert(gearB, format(L["%d |4item:items;"], brokenSlots))
			elseif lowSlots > 0 then
				if status ~= "red" then status = "yellow" end
			 table.insert(gearA, colorize(L["Low durability"], yellow))
			 table.insert(gearB, format(L["%d |4item:items;"], lowSlots))
			end
			if missingSlots > 0 then
				status = "red"
			 table.insert(gearA, colorize(L["Not equipped"], red))
			 table.insert(gearB, format(L["%d |4item:items;"], missingSlots))
			end
			
			-- Check hunter ammo
			if E.class == "HUNTER" then
				local ammoSlot = GetInventorySlotInfo("AmmoSlot")
				local ammoCount = GetInventoryItemCount("player", ammoSlot)
				if ammoCount == 1 and not GetInventoryItemTexture("player", ammoSlot) then ammoCount = 0 end
				if ammoCount <= 0 then
					status = "red"
					table.insert(gearA, colorize(L["Expected"], red))
					table.insert(gearB, L["Ammunition"])
				elseif ammoCount < self.db.profile.minAmmo then
					if status ~= "red" then status = "yellow" end
					table.insert(gearA, colorize(L["Low"], yellow))
					table.insert(gearB, L["Ammunition"])
				end
			end
			
			-- Count issues and set icon
			numIssues = numIssues + #otherIssuesA.me + #otherIssuesA.pet + #gearA
			if #otherIssuesA.me > 0 or #otherIssuesA.pet > 0 then status = "red" end
			self:SetStatus(status)
			
			-- Set text
			local text = ""
			if numIssues > 0 then
				if self.db.profile.feedText == 3 then
					text = format(L["%d |4issue:issues;"], numIssues)
				elseif self.db.profile.feedText == 2 then
					text = format("%d", numIssues)
				end
			end
			self:SetText(text)
			
			-- Set the feed to auto hide if we're good
			if self.db.profile.hideTime > 0 and status == "green" and not hideScheduled and not self:InTooltip() and not isHidden and not hideTimer then
				hideScheduled = true
				hideTimer = self:ScheduleTimer("OnHideTimer", self.db.profile.hideTime)
			elseif status ~= "green" then
				hideScheduled = false
				self:CancelTimer(hideTimer, true)
				self:ShowFeed()
			end
			
			-- Update the tooltip if it's currently showing
			if self:InTooltip() then E:ShowTooltip() end
		end
	end
	
	
	function E:OnHideTimer()
		if hideScheduled and not self:InTooltip() and not isHidden then
			self:HideFeed()
			hideScheduled = false
			hideTimer = nil
		end
	end


	function E:Update(event)
		if event then updateSource = event end
		if (lastUpdate == nil or GetTime() - lastUpdate >= UPDATE_THROTTLE) and not updateScheduled then
			self:_Update()
		else
			if not updateScheduled then
				updateScheduled = true
				self:ScheduleTimer("_Update", UPDATE_THROTTLE + 0.1)
			end
		end
	end
end


--[[
  
  Status types have hardcoded definitions.
  
  Expected and Unexpected: Error
  Low and Missing: Warning
  
]]

function E:EvaluateAllGroups()
	local status = "green"
	local numIssues = 0
	for _, group in pairs(self.db.profile.groups) do
		if not group.hidden then
			self:EvaluateGroup(group)
			if group._result == "Low" and self.db.profile.lowWarn then
				if status == "green" then status = "yellow" end
				numIssues = numIssues + 1
			elseif group._result == "Expected" or group._result == "Unexpected" then
				status = "red"
				numIssues = numIssues + 1
			end
			if group._result_pet == "Low" and self.db.profile.lowWarn and self.db.profile.petBuffs and E.isPetClass and UnitExists("playerpet") then
				if status == "green" then status = "yellow" end
				numIssues = numIssues + 1
			elseif (group._result_pet == "Expected" or group._result_pet == "Unexpected") and self.db.profile.petBuffs and E.isPetClass and UnitExists("playerpet") then
				status = "red"
				numIssues = numIssues + 1
			elseif group._result == "Missing" then
				if status == "green" then status = "yellow" end
				numIssues = numIssues + 1
			end
		end
	end
	return status, numIssues
end


function E:EvaluateGroup(group)
	group._status = "not_scanned"
	group._status_pet = "not_scanned"
	
	for _, item in pairs(group.items) do
		self:EvaluateItem(item)
		if item._status == "low" and group._status ~= "found" then group._status = "low"
		elseif item._status == "found" then group._status = "found"
		elseif item._status == "not_found" and group._status ~= "found" and group._status ~= "low" then group._status = "not_found"
		end
		if item._status_pet == "low" and group._status_pet ~= "found" then group._status_pet = "low"
		elseif item._status_pet == "found" then group._status_pet = "found"
		elseif item._status_pet == "not_found" and group._status_pet ~= "found" and group._status_pet ~= "low" then group._status_pet = "not_found"
		end
	end
	
	local foundbuff = false
	local foundconsumable = false
	local founditem = false
	local failedconsumables = 0
	local failedconsumables_pet = 0
	local scannedbuffs = 0
	local scannedbuffs_pet = 0
	local faileditems = 0
	
	for _, item in pairs(group.items) do
		if item.type == "BUFF" then
			foundbuff = true
			if item._status ~= "not_scanned" then scannedbuffs = scannedbuffs + 1 end
			if item._status_pet ~= "not_scanned" then scannedbuffs_pet = scannedbuffs_pet + 1 end
		elseif item.type == "CONSUMABLE" then
			foundconsumable = true
			if item._status == "not_found" then failedconsumables = failedconsumables + 1 end
			if item._status_pet == "not_found" then failedconsumables_pet = failedconsumables_pet + 1 end
		elseif item.type == "ITEM" then
			founditem = true
			if item._status == "low" then faileditems = faileditems + 1 end
		end
	end
	if foundbuff and foundconsumable and scannedbuffs <= 0 and failedconsumables > 0 then group._status = "not_scanned" end
	if foundbuff and foundconsumable and scannedbuffs_pet <= 0 and failedconsumables_pet > 0 then group._status_pet = "not_scanned" end
	
	group._result = nil
	group._result_pet = nil
	if group._status ~= "not_scanned" then
		if founditem and faileditems > 0 then group._result = "Missing"
		elseif group._status == "not_found" and group.me == "EXPECTED" then group._result = "Expected"
		elseif group._status ~= "not_found" and group.me == "UNEXPECTED" then group._result = "Unexpected"
		elseif group._status == "low" and group.me == "EXPECTED" then group._result = "Low"
		end
	end
	if group._status_pet ~= "not_scanned" then
		if group._status_pet == "not_found" and group.pet == "EXPECTED" then group._result_pet = "Expected"
		elseif group._status_pet ~= "not_found" and group.pet == "UNEXPECTED" then group._result_pet = "Unexpected"
		elseif group._status_pet == "low" and group.pet == "EXPECTED" then group._result_pet = "Low"
		end
	end
end


function E:EvalList(item, doMe, doPet, fn, ...)
	local v, val
	for i = 1, select("#", ...) do
		v = select(i, ...)
		val = strupper(strtrim(v))
		if val and val ~= "" then
			self[fn](self, item, doMe, doPet, val)
		end
	end
end


function E:EvaluateItem_BUFF_OR_CONSUMABLE(item, doMe, doPet, s)
	-- Name matching using ID
	if strstarts(s, "\"") and strends(s, "\"") then
		s = strsub(s, 2, strlen(s) - 1)
		id = tonumber(s)
		if id then
			s = GetSpellInfo(id)
			if s then
				s = strupper(s)
				if doMe then
					for i, v in pairs(aura.me.spellName) do
						if compare_funcs[item.mode](v, s) then
							item._status = "found"
							if aura.me.unitCaster[i] and UnitExists(aura.me.unitCaster[i]) then
								item._lastProvidedBy = UnitName(aura.me.unitCaster[i])
							end
							item._lastSeen = GetTime()
							if aura.me.expiration[i] <= item.warn and aura.me.expiration[i] > 0 and self.db.profile.lowWarn then item._status = "low" end
							break
						end
					end
				end
				if doPet then
					for i, v in pairs(aura.pet.spellName) do
						if compare_funcs[item.mode](v, s) then
							item._status_pet = "found"
							if aura.pet.unitCaster[i] and UnitExists(aura.pet.unitCaster[i]) then
								item._lastProvidedBy_pet = UnitName(aura.pet.unitCaster[i])
							end
							item._lastSeen_pet = GetTime()
							if aura.pet.expiration[i] <= item.warn and aura.pet.expiration[i] > 0 and self.db.profile.lowWarn then item._status_pet = "low" end
							break
						end
					end
				end
			end
		end
		
	-- ID matching
	elseif tonumber(s) then
		id = tonumber(s)
		if doMe then
			for i, v in pairs(aura.me.spellId) do
				if v == id then
					item._status = "found"
					if aura.me.unitCaster[i] and UnitExists(aura.me.unitCaster[i]) then
						item._lastProvidedBy = UnitName(aura.me.unitCaster[i])
					end
					item._lastSeen = GetTime()
					if aura.me.expiration[i] <= item.warn and aura.me.expiration[i] > 0 and self.db.profile.lowWarn then item._status = "low" end
					break
				end
			end
		end
		if doPet then
			for i, v in pairs(aura.pet.spellName) do
				if v == id then
					item._status_pet = "found"
					if aura.pet.unitCaster[i] and UnitExists(aura.pet.unitCaster[i]) then
						item._lastProvidedBy_pet = UnitName(aura.pet.unitCaster[i])
					end
					item._lastSeen_pet = GetTime()
					if aura.pet.expiration[i] <= item.warn and aura.pet.expiration[i] > 0 and self.db.profile.lowWarn then item._status_pet = "low" end
					break
				end
			end
		end
		
	-- Name matching
	else
		if doMe then
			for i, v in pairs(aura.me.spellName) do
				if compare_funcs[item.mode](v, s) then
					item._status = "found"
					if aura.me.unitCaster[i] and UnitExists(aura.me.unitCaster[i]) then
						item._lastProvidedBy = UnitName(aura.me.unitCaster[i])
					end
					item._lastSeen = GetTime()
					if aura.me.expiration[i] <= item.warn and aura.me.expiration[i] > 0 and self.db.profile.lowWarn then item._status = "low" end
					break
				end
			end
		end
		if doPet then
			for i, v in pairs(aura.pet.spellName) do
				if compare_funcs[item.mode](v, s) then
					item._status_pet = "found"
					if aura.pet.unitCaster[i] and UnitExists(aura.pet.unitCaster[i]) then
						item._lastProvidedBy_pet = UnitName(aura.pet.unitCaster[i])
					end
					item._lastSeen_pet = GetTime()
					if aura.pet.expiration[i] <= item.warn and aura.pet.expiration[i] > 0 and self.db.profile.lowWarn then item._status_pet = "low" end
					break
				end
			end
		end
	end
end


function E:EvaluateItem_ITEM(item, doMe, doPet, s)
	-- Name matching using ID
	if strstarts(s, "\"") and strends(s, "\"") then
		s = strsub(s, 2, strlen(s) - 1)
		id = tonumber(s)
		if id then
			s = GetItemInfo(id)
			if s then
				s = strupper(s)
				for i, v in pairs(items) do
					if compare_funcs[item.mode](strupper(GetItemInfo(i)) or "", s) then
						if v >= item.quantity then
							item._status = "found"
							item._lastSeen = GetTime()
						end
						break
					end
				end
			end
		end
		
	-- ID matching
	elseif tonumber(s) then
		id = tonumber(s)
		for i, v in pairs(items) do
			if i == id then
				if v >= item.quantity then
					item._status = "found"
					item._lastSeen = GetTime()
				end
				break
			end
		end
		
	-- Name matching
	else
		for i, v in pairs(items) do
			if compare_funcs[item.mode](strupper(GetItemInfo(i)) or "", s) then
				if v >= item.quantity then
					item._status = "found"
					item._lastSeen = GetTime()
				end
				break
			end
		end
	
	end
end


function E:EvaluateItem(item)
	local i, id, v
	item._status = "not_scanned"
	item._status_pet = "not_scanned"
	
	-- Hidden
	if not item.hidden then
	
		-- Class present
		if (classes[item.class] or 0) > 0 or item.type ~= "BUFF" then
	
			-- Instance difficulty
			if item.difficulty == "ANY" or (item.difficulty == "HEROIC" and isHeroic) or (item.difficulty == "NORMAL" and not isHeroic) then
			
				-- Party format
				if item.partyFormat == "ANY" or (item.partyFormat == "ANY_RAID" and UnitInRaid("player")) or item.partyFormat == partyFormat then

					-- Talent
					if self:RaidHasTalent(item.class, item.talentTree, item.talentIndex) then
					
-- BUFF or CONSUMABLE ------------------------------------------------------------------------------------------------------------------------

						if item.type == "BUFF" or item.type == "CONSUMABLE" then
						
							item._status = "not_found"
							item._status_pet = "not_found"
							
							local doMe = true
							local doPet = true
							local prio = tonumber(item.priorityMe)
							if prio and prio > 0 and classes[item.class] < prio then
								doMe = false
								item._status = "not_scanned"
							end
							prio = tonumber(item.priorityPet)
							if prio and prio > 0 and classes[item.class] < prio then
								doPet = false
								item._status_pet = "not_scanned"
							end
							
							self:EvalList(item, doMe, doPet, "EvaluateItem_BUFF_OR_CONSUMABLE", strsplit(",", item.value))
						
-- ITEM --------------------------------------------------------------------------------------------------------------------------------------

						elseif item.type == "ITEM" then

							if self.db.profile.itemWarn then
								item._status = "low"
								item._status_pet = "not_scanned"
								self:EvalList(item, true, true, "EvaluateItem_ITEM", strsplit(",", item.value))
							else
								item._status = "not_scanned"
								item._status_pet = "not_scanned"
							end

						elseif item.type == "ON_OTHERS" then
						end
					
					end -- talent
				end -- party format
			end -- difficulty
		end -- class present
	end -- hidden
end


function E:RaidHasTalent(class, talentTree, talentIndex)
	if not talentTree or talentTree <= 0 or not talentIndex or talentIndex <= 0 then return true end
	for k, v in pairs(players) do
		if v == class then
			if (select(5, talents:GetTalentInfo(k, talentTree, talentIndex)) or 0) > 0 then return true end
		end
	end
	return false
end


function E:ShouldEnable()
	if UnitInRaid("player") then
		for i = 1, MAX_BATTLEFIELD_QUEUES do
			local status, mapName, instanceID, minlevel, maxlevel, teamSize = GetBattlefieldStatus(i)
			if status and status == "active" then
				if teamSize and teamSize > 0 then
					return self.db.profile.enableArena
				else
					return self.db.profile.enableBattleground
				end
			end
		end
		return self.db.profile.enableRaid
	elseif GetNumPartyMembers() > 0 then
		return self.db.profile.enableParty
	else
		return self.db.profile.enableSolo
	end
end


function E:CountClasses()
	for _, v in pairs(CLASS_LIST) do
		classes[v] = 0
	end
	for _, v in pairs(players) do
		classes[v] = classes[v] + 1
	end
end


do
	local lastIcon, lastText

	function E:SetStatus(icon, text)
		lastIcon = icon
		if isHidden then icon = "blank" end
		if icon ~= "blank" then
			icon = [[Interface\AddOns\Energized\Icons\]]..(self.db.profile.iconSet or "Default").."\\"..icon
		else
			icon = [[Interface\AddOns\Energized\Icons\]]..icon
		end
		dataobj.icon = icon
		if text then
			lastText = text
			if isHidden then text = "" end
			dataobj.text = text
		end
	end


	function E:SetText(text)
		lastText = text or ""
		if isHidden then text = "" end
		dataobj.text = text or ""
	end
	
	
	function E:HideFeed()
		if not isHidden then
			self:Debug("Hiding feed")
			isHidden = true
			self:SetStatus(lastIcon, lastText)
		end
	end
	
	
	function E:ShowFeed()
		if isHidden then
			self:Debug("Showing feed")
			isHidden = false
			self:SetStatus(lastIcon, lastText)
		end
	end
end


function E:GetPetMode()
	for i = 1, 10 do
		local name, subtext, texture, isToken, isActive, autoCastAllowed, autoCastEnabled = GetPetActionInfo(i)
		if name == "PET_MODE_AGGRESSIVE" and isActive == 1 then return "PET_MODE_AGGRESSIVE"
		elseif name == "PET_MODE_DEFENSIVE" and isActive == 1 then return "PET_MODE_DEFENSIVE"
		elseif name == "PET_MODE_PASSIVE" and isActive == 1 then return "PET_MODE_PASSIVE"
		end
	end
	return "PET_MODE_UNKNOWN"
end


function E:FindLastBufferMe(group)
	local time = 0
	local by = ""
	for _, item in pairs(group.items) do
		if item._lastProvidedBy and item._lastSeen then
			if time < item._lastSeen then
				time = item._lastSeen
				by = item._lastProvidedBy
			end
		end
	end
	return by, time
end


function E:FindLastBufferPet(group)
	local time = 0
	local by = ""
	for _, item in pairs(group.items) do
		if item._lastProvidedBy_pet and item._lastSeen_pet then
			if time < item._lastSeen_pet then
				time = item._lastSeen_pet
				by = item._lastProvidedBy_pet
			end
		end
	end
	return by, time
end
